/* $Header: reg2nfa.cv  1.10  91/06/07 10:53:56  Exp $ */
/* from Regular Expressions to Nondeterministic Finite Automata */

#include <stdio.h>
#include <stdlib.h>
#include <strings.h>

#include "reg.h"

extern char used[CHAR_STAFF];

void FatalError(char *, int);

void cutR(char *, nfa_state_ptr *, nfa_state_ptr *);
void Join(nfa_state_ptr, nfa_state_ptr, nfa_state_ptr, nfa_state_ptr, nfa_state_ptr *, nfa_state_ptr *);
void Quantifier(nfa_state_ptr, nfa_state_ptr, nfa_state_ptr *, nfa_state_ptr *);
void Charactor(char, nfa_state_ptr *, nfa_state_ptr *);
void Alternation(nfa_state_ptr, nfa_state_ptr, nfa_state_ptr, nfa_state_ptr, nfa_state_ptr *, nfa_state_ptr *);

/* 正規表現からNFAを作り出す */
nfa_state_ptr reg2nfa(char *reg)
{
	nfa_state_ptr i, f;
	nfa_state_ptr MakeNewNfaState();

	i = f = NULL;
	cutR(reg, &i, &f); /* 正規表現を構成している要素を分解 */
	if (i == NULL) {
		/* 対象になる式が見つからなければ、空のNFAを作り出す */
		i = MakeNewNfaState();
		f = MakeNewNfaState();
		i->input[0] = EPSILON;
		i->transition[0] = f;
	}
	f->input[0] = GOAL;
	return (i);
}


/* 正規表現を記述するために必要な記号かどうか調べる */
int NotMetaChara(char c)
{
	switch (c) {
	 case '(' :
	 case ')' :
	 case '*' :
	 case '|' :
	 	return (FALSE);
	 default :
	 	return (TRUE);
	}
}



/* 式を分解して、それぞれのNFAを作る。
 * 後にできあがったNFAを合成して、全体のNFAとする。
 */
void cutR(char *str, nfa_state_ptr *i, nfa_state_ptr *f)
{
	nfa_state_ptr i0, f0, i1, f1;
	int pare = 0;
	int ip;
	char *p;
	char this[128];

	i0 = f0 = i1 = f1 = NULL;

	for (p = str; ; ) { /* 式を一文字ずつ順に読み込む */

		switch (*p) { /* 読み込んだ文字によって分類する */

		 case '\0' : /* 文字列の終り */
		 	Join(i0, f0, i1, f1, i, f); /* 作ったNFAを連結 */
		 	return;

		 case '(' : /* 括弧の始まり */
		 	p++;
			for (pare = 1, ip = 0; pare != 0; ip++, p++) {
				switch (*p) {

				 case '\0' : /* 括弧閉じがないのでエラー */
					FatalError("pare",101);

				 case '(' :
				 	pare++;
				 	this[ip] = *p;
				 	break;

				 case ')' :
				 	/* 括弧閉じが来たので一つのNFA
				 	 * として処理する */
				 	pare--;
				 	if (pare == 0) {
				 		this[ip] = '\0';
				 		Join(i0, f0, i1, f1, &i0, &f0);
				 		cutR(this, &i1, &f1);
					} else
						this[ip] = *p;
					break;

				 default :
				 	this[ip] = *p;
				 	break;
				}
			}
			break;

		 case '*' : /* Star */
		 	Quantifier(i1, f1, &i1, &f1);
		 	p++;
		 	break;
		
		 case '|' : /* Or */
		 	Join(i0, f0, i1, f1, &i0, &f0);
		 	p++;
		 	cutR(p, &i1, &f1); /* 後ろに続く式は別に評価 */
		 	Alternation(i0, f0, i1, f1, i, f);
		 	return;

		 default : /* 普通の文字 */
		 	Join(i0, f0, i1, f1, &i0, &f0);
		 	Charactor(*p, &i1, &f1);
		 	p++;
		 	break;
		}
	}

}


/* 2つのNFAの合成 */
void Join(nfa_state_ptr i1, nfa_state_ptr f1, nfa_state_ptr i2, nfa_state_ptr f2, nfa_state_ptr *i, nfa_state_ptr *f)
{
	if (i1 == NULL) {
		*i = i2;
		*f = f2;
	} else {
		*i = i1;
		*f = f2;
		bcopy(i2, f1, sizeof(struct nfa_state));
		//memcpy(i2, f1, sizeof(struct nfa_state));
		free(i2);
	}
}


/* Quantifier * について */
void Quantifier(nfa_state_ptr i, nfa_state_ptr f, nfa_state_ptr *ri, nfa_state_ptr *rf)
{
	nfa_state_ptr MakeNewNfaState();
	nfa_state_ptr ti, tf;

#ifndef OLD  /* これはバグ付き */
	if (i == NULL) {
		i = MakeNewNfaState();
		f = MakeNewNfaState();
		i->input[0] = EPSILON;
		i->transition[0] = f;
		f->input[0] = EPSILON;
	}
	ti = MakeNewNfaState();
	tf = MakeNewNfaState();
	ti->input[0] = EPSILON;
	ti->transition[0] = i;
	ti->input[1] = EPSILON;
	ti->transition[1] = tf;
	tf->input[0] = EPSILON;
	f->transition[0] = tf;
	f->input[1] = EPSILON;
	f->transition[1] = i;
	*ri = ti;
	*rf = tf;
#else /* これが正式 */
/* 下のグラフを作る
         ε
       +----+
       |    |
    ε v a  | ε
 ri -> i -> f -> rf
  |              ^
  +--------------+
         ε
*/
	if (i != NULL) {
		ti = MakeNewNfaState();
		tf = MakeNewNfaState();
		ti->input[0] = EPSILON;
		ti->transition[0] = i;
		ti->input[1] = EPSILON;
		ti->transition[1] = tf;
		tf->input[0] = EPSILON;
		f->transition[0] = tf;
		f->input[1] = EPSILON;
		f->transition[1] = i;
		*ri = ti;
		*rf = tf;
	} else {
		*ri = *rf = NULL;
	}
#endif
}


/* 文字の処理 */
void Charactor(char c, nfa_state_ptr *i, nfa_state_ptr *f)
{
/*下のグラフを作る
      c
  i  --->  f
 */
	nfa_state_ptr MakeNewNfaState();

	*i = MakeNewNfaState();
	*f = MakeNewNfaState();
	(*i)->input[0] = c;
	(*i)->transition[0] = *f;
	(*f)->input[0] = EPSILON;
	used[0x00ff & (int)c] = '1';
}


/* Alternation | の処理 */
void Alternation(nfa_state_ptr i1, nfa_state_ptr f1, nfa_state_ptr i2, nfa_state_ptr f2, nfa_state_ptr *i, nfa_state_ptr *f)
{
/* 下のグラフを作る
         i1 -> f1
       /         `
     i             f
       `         /
         i2 -> f2
 */
	nfa_state_ptr MakeNewNfaState();
	nfa_state_ptr ti, tf;

	ti = MakeNewNfaState();
	tf = MakeNewNfaState();

	ti->input[0] = EPSILON;
	ti->input[1] = EPSILON;
	if (i1 != NULL) {
		ti->transition[0] = i1;
		f1->input[0] = EPSILON;
		f1->transition[0] = tf;
	} else
		ti->transition[0] = tf;

	if (i2 != NULL) {
		ti->transition[1] = i2;
		f2->input[0] = EPSILON;
		f2->transition[0] = tf;
	} else
		ti->transition[1] = tf;

	tf->input[0] = EPSILON;
	*i = ti;
	*f = tf;
}


/* NFAのために新しい領域を割り当てる */
nfa_state_ptr MakeNewNfaState()
{
	nfa_state_ptr new;

	if ((new = (nfa_state_ptr)malloc(sizeof(nfa_state))) == NULL)
		FatalError("malloc",100);
	bzero(new, sizeof(nfa_state));
	//memset(new, 0, sizeof(nfa_state));
	return (new);
}
